SLRN_HAS_PULL_SUPPORT		1
SLRN_HAS_NNTP_SUPPORT		1
SLRN_HAS_SPOOL_SUPPORT		1
SLRN_HAS_INEWS_SUPPORT		0
SLRN_FORCE_INEWS		0
SLRN_INEWS_COMMAND		inews
SLRN_HAS_UUDEVIEW		0
SLRN_HAS_CANLOCK		0
SLRN_HAS_STRICT_FROM		0
SLRN_HAS_CHARACTER_MAP		1
SLRNPULL_USE_SETGID_POSTS	0
SLRN_HAS_SPOILERS		1
SLRN_HAS_EMPHASIZED_TEXT	1
SLRN_HAS_GEN_MSGID		1
SLRN_HAS_MSGID_CACHE		1
SLRN_HAS_FAKE_REFS		1
SLRN_HAS_GROUPLENS		0
