#error This file has been removed from the slrn source tarball because its
#error license is more restrictive than the one of slrn itself.

#error If you want to compile slrn for VMS and need this file, please get it
#error from http://slrn.sourceforge.net/data/clientlib.tar.gz
